@@ -7,6 +7,9 @@ import snowflake.connector
 from tinybird_cdk import (cloud, config, connector, errors, export, formats, schema,
                           logger, utils)
 
+
+LOGIN_TIMEOUT_IN_SECONDS = 10
+
 # Snowflake specific classes.
 # This is an abstraction leak, for Snowflake specific methods.
 @dataclass(frozen=True)
@@ -19,6 +22,11 @@ class Warehouse:
     name: str
     size: str
 
+@dataclass(frozen=True)
+class Integration:
+    name: str
+    type: str
+
 # The `snowql` CLI for Snowflake understands a good deal of environment
 # variables:
 #
@@ -42,6 +50,12 @@ class Connector(connector.SQLConnector):
     def __init__(self):
         super().__init__()
 
+    # Helper to get a property from a Snowflake `describe` resultset
+    def _get_property(self, object_properties, property_name):
+        property = list(filter(lambda p: p['property'] == property_name, object_properties))
+        if len(property):
+            return property[0]['property_value']
+
     # Get available roles for the provided user.
     # Tinybird needs to present the user with a list of available roles to configure the connection in the UI.
     # These two methods (get_roles, get_warehouses) are only avaiable in the Snowflake Connector, this is effectively an abstraction leak
@@ -62,6 +76,14 @@ class Connector(connector.SQLConnector):
         warehouses.sort(key=lambda warehouse: warehouse.name)
         return warehouses
 
+    # Get available integrations for the provided user and role
+    def get_integrations(self):
+        integrations = []
+        for integration in list(self._query('show integrations')):
+            integrations.append(Integration(name=integration['name'], type=integration['type']))
+        integrations.sort(key=lambda integration: integration.name)
+        return integrations
+
     # Create or replace Snowflake integration and stage.
     # https://docs.snowflake.com/en/sql-reference/sql/create-storage-integration
     # https://docs.snowflake.com/en/sql-reference/sql/create-stage
@@ -78,34 +100,79 @@ class Connector(connector.SQLConnector):
     #
     # Returns the stage name, and the service account Snowflake generates for the integration. Tinybird Analytics
     # has to grant this account access to the bucket.
-    def create_stage(self, allowed_location):
-        integration_name = 'tinybird_integration'
-        stage_name = 'tinybird_stage'
-
+    def create_stage(self, allowed_location , integration_name='tinybird_integration'):
         if not (database := config.get('SF_DATABASE')):
             raise errors.MissingConfiguration('database')
         if not (schema := config.get('SF_SCHEMA')):
             raise errors.MissingConfiguration('schema')
-        
+        if not (workspace_id := config.get('TB_WORKSPACE_ID')):
+            raise errors.MissingConfiguration('workspace_id')
+
+
+        # Recreating a storage integration (using CREATE OR REPLACE STORAGE INTEGRATION) breaks the association
+        # between the storage integration and any stage that references it.
+        #
+        # This is because a stage links to a storage integration using a hidden ID rather than the name of
+        # the storage integration.
+        #
+        # Behind the scenes, the CREATE OR REPLACE syntax drops the object and recreates it with a different hidden ID.
+        # https://docs.snowflake.com/en/sql-reference/sql/create-storage-integration#usage-notes
+        # 
+        # That's why we're only creating the integration once
+        # 
+        # NOTE: Commented out, since we're requiring the user to create the integration on their end.
+        #
+        # self._query(f'''
+        #     create storage integration if not exists "{integration_name}"
+        #       type = external_stage
+        #       storage_provider = 'GCS'
+        #       enabled = true
+        #       comment = 'Tinybird Snowflake Connector Integration'
+        #       storage_allowed_locations = ('{allowed_location}')
+        # ''')
+
+        # Since the integration is only created once, we need to update the STORAGE_ALLOWED_LOCATIONS property
+        # and append new workspace folder if they do not exist already.
+        integration = self._query(f''' describe integration "{integration_name}" ''');
+        allowed_locations = self._get_property(integration, 'STORAGE_ALLOWED_LOCATIONS').split(",")
+        if allowed_location not in allowed_locations:
+            allowed_locations.append(allowed_location)
+            self._query(f'''
+                alter storage integration if exists "{integration_name}" set
+                    storage_allowed_locations = ({', '.join(f'"{l}"' for l in allowed_locations)});
+            ''')
+
+        # Create one stage per workspace, since every stage must include the allowed location URL with
+        # its workpace id in the path.
+        #
+        # Replace "-" chars for "_", so we can avoid using quotes in Snowflake resources and complicate everything
+        stage_name = f'{database}.{schema}.tinybird_stage_{workspace_id.replace("-", "_")}'
         self._query(f'''
-            create or replace storage integration {integration_name}
-              type = external_stage
-              storage_provider = 'GCS'
-              enabled = true
-              storage_allowed_locations = ('{allowed_location}')
-        ''')
-        
-        self._query(f'''
-            create or replace stage {database}.{schema}.{stage_name}
+            create or replace stage {stage_name}
               url='{allowed_location}'
               file_format = (type = csv compression = none )
-              storage_integration = {integration_name};
+              storage_integration = "{integration_name}";
         ''')
 
-        integration_properties = self._query(f'describe integration {integration_name}')
-        account = list(filter(lambda p: p['property'] == 'STORAGE_GCP_SERVICE_ACCOUNT', integration_properties))[0]['property_value']
+        account = self._get_property(integration, 'STORAGE_GCP_SERVICE_ACCOUNT')
 
-        return { "stage": f'{database}.{schema}.{stage_name}', "gcp_account": account }
+        return { "stage": stage_name, "gcp_account": account }
+
+    def get_integration_query(self, allowed_location, integration_name='tinybird_integration'):
+        if not (role := config.get('SF_ROLE')):
+            raise errors.MissingConfiguration('sf_role')
+
+        # Remove tabs
+        return re.sub(r"( )+", " ", f'''
+            create storage integration if not exists "{integration_name}"
+                \ttype = external_stage
+                \tstorage_provider = 'GCS'
+                \tenabled = true
+                \tcomment = 'Tinybird Snowflake Connector Integration'
+                \tstorage_allowed_locations = ('{allowed_location}');
+
+            grant ownership on integration "{integration_name}" to role {role}; 
+        ''')
 
     #
     # --- Scopes ----------------------------------------------------------------------------------
@@ -131,7 +198,7 @@ class Connector(connector.SQLConnector):
 
         # The 'show databases' sentece includes also default databases SNOWFLAKE and SNOWFLAKE_SAMPLE_DATA.
         # Apparently, there's no way to filter out those databases from the result set. So we'll filter them manually.
-        filtered_databases =  list(filter(lambda d: d['name'] not in ('SNOWFLAKE', 'SNOWFLAKE_SAMPLE_DATA'), raw_databases))
+        filtered_databases = list(filter(lambda d: d['name'] not in ('SNOWFLAKE', 'SNOWFLAKE_SAMPLE_DATA'), raw_databases))
 
         for database in filtered_databases:
             databases.append(connector.Scope(name=database['name'], value=database['name']))
@@ -292,5 +359,6 @@ class Connector(connector.SQLConnector):
             role=config.get('SF_ROLE'),
             warehouse=config.get('SF_WAREHOUSE'),
             database=config.get('SF_DATABASE'),
-            schema=config.get('SF_SCHEMA')
+            schema=config.get('SF_SCHEMA'),
+            login_timeout=LOGIN_TIMEOUT_IN_SECONDS
         )
